// VinciaColourReconnections.h is a part of the VINCIA plug-in to the PYTHIA 8 generator.
// Copyright (C) 2014 P. Skands.
// VINCIA is licenced under the GNU GPL version 2.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// This file contains header information for VINCIA's ColourReconnections class

#ifndef VINCIACOLOURRECONNECTIONS_H
#define VINCIACOLOURRECONNECTIONS_H

#include "VinciaBranchelemental.h"   // Common Vincia functions
#include "VinciaCommon.h"   // Common Vincia functions
#include "math.h"

namespace Vincia {
  
  // *****************************************************************
  // class ColourReconnections
  // for (perturbative and non-perturbative) colour reconnections
  // Implementation is similar (but not identical) to idea of "dipole
  // swing" in ARIADNE. 

  class ColourReconnections {

  public:
      
    // Constructor.
    
    ColourReconnections() {isInitPtr=false; isInit=false;}

    // Destructor.
    ~ColourReconnections() {}

    // Initialize Pointers (must be done before init)
    void initPtr(Rndm* rndmPtrIn, Settings* settingsPtrIn) {
      rndmPtr        = rndmPtrIn;
      settingsPtr    = settingsPtrIn; 
      isInitPtr=true;
    }
    
    // Initialize
    bool init();

    // Get translation map from ordinary Les Houches tags to antenna indices 
    // (tags always ending on [1-9] with gluons having non-identical colour 
    //  and anticolour indices). Start new tags off from startTag.
    // Return true if any map is nontrivial (i.e., if there are any 
    // remappings to be done), otherwise false.
    bool getColourMap(map<int,int>& colorMap,  
                      vector<Particle>& particles, int startTag);

    // Reconnect a system of antennae, allowing reconnections between
    // invariants down to the scale q2now (by default zero, for non-
    // perturbative reconnections, for which all invariants are allowed
    // to participate). The measureType selects between different possible
    // measures used to compare the antennae. 
    // Type: 1   : Lambda product, with new lambdas > q2now 
    //       2   : Lambda product, with new lambdas > q2now and 
    //             both lambda values decreasing.
    int doReconnect(Event& event, vector<Branchelemental>& antPtrs,
                    double q2now = 0.0, int measureType = 1);

    // Look for reconnections that could produce junctions
    int doJunctions(Event& event, vector<Branchelemental>& antennae);

    // String length measure
    double lambda(Vec4 p1, Vec4 p2) {
      return 2. * (dot4(p1,p2) - m(p1)*m(p2)) + m2min;
    }

    // Set verbose level
    void setVerbose(int verboseIn) {verbose = verboseIn;};

    // Print header information
    void header(int len=9);

  private:

    // Internal parameters
    int CRmode;
    double mMin, m2min, junctionsM0;
    bool CRjunctions;

    // Is initialized?    
    bool isInitPtr, isInit;
    
    // Pointer to PYTHIA 8 Random-number generator and Settings database
    Rndm*     rndmPtr;
    Settings* settingsPtr;

    // Verbose level
    int verbose;   

  };

//**************************************************************************
  
  // A simple matrix class designed for colour-reconnection algebra
  
  class CRmatrix {

  public: 

    // Constructor
    CRmatrix() {};
    
    // Functions to get
    double getEntry(int iRow, int iCol) {
      if (iCol >= int(columns.size())) return 0.0;
      vector<double> column = columns[columnMap[iCol]];
      if (iRow >= int(column.size())) return 0.0;
      return column[iRow];
    }
    vector<double> getColumn(int iCol) {
      return columns[columnMap[iCol]];
    }
    vector<double> getDiagonal() {
      vector<double> diagonal;
      for (int i=0; i<int(columns.size()); ++i) 
        diagonal.push_back(getEntry(i,i));
      return diagonal;
    }
    // Get the index of the n'th largest entry on the diagonal
    int getDiagonalIndexOrdered(int rank) {
      if (rank >= int(columns.size()) || rank < 0) return -1;
      vector<double> diagonal = getDiagonal();
      vector<double> diagonalSorted(diagonal);
      sort(diagonalSorted.begin(),diagonalSorted.end());
      double val = diagonalSorted[diagonal.size()-1-rank];
      for (int i=0; i<int(diagonal.size()); ++i) 
        if (diagonal[i] == val) return i;
      return -1;
    }
    // Get the index corresponding to the new column (if swaps done)
    int getNewColumn(int iCol) {
      return columnMap[iCol];
    }
      
    // Functions to set
    void setEntry(int iRow, int iCol, double val) {
      if (iCol >= int(columns.size())) {
        columns.resize(iCol+1);
        columnMap[iCol] = iCol;
      }
      vector<double> column = columns[columnMap[iCol]];
      if (iRow >= int(column.size())) column.resize(iRow+1);
      column[iRow] = val;
      columns[columnMap[iCol]] = column;
    }

    // Swap two columns
    void swapColumns(int iCol, int jCol) {
      if (max(iCol,jCol) >= int(columns.size())) return;
      columnMap[iCol] = jCol;
      columnMap[jCol] = iCol;
    }

  private:
    vector< vector<double> > columns;
    map<int, int> columnMap;
  };

}

#endif
