// VinciaCommon.h is a part of the VINCIA plug-in to the PYTHIA 8 generator.
// Copyright (C) 2014 P. Skands.
// VINCIA is licenced under the GNU GPL version 2.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// This file contains header information for the auxiliary helper 
// functions for VINCIA, including the VinciaCommon and AlphaQCD classes, 
// the latter of which inherits from Pythia8::AlphaStrong, but adds a bit 
// of extra functionality.

#ifndef VINCIACOMMON_H
#define VINCIACOMMON_H

// Include Pythia 8 headers and name space

#include "Pythia8/Event.h"           // Event
#include "Pythia8/StandardModel.h"   // AlphaStrong

using namespace Pythia8;

// ------------ GLOBAL VINCIA CONSTANTS AND FUNCTIONS --------------------
namespace Vincia {

  // Global numerical precision target
  // (should be kept within a reasonable margin of machine precision)
  // (large values may result in strange results due to non-zero terms
  //  being artificially forced to zero)
  const double TINY    = 1.0e-9;
  
  //**************************************************************************

  // Utilities for printing out VINCIA Messages
  void printErr(string,string);
  void printOut(string,string);
   
  // Utilities for converting data to strings
  string num2str(int,int width=4) ;
  string num2str(double,int width=9) ;
  string bool2str(bool, int width=2) ;
  
  //**************************************************************************

  // VinciaCommon: 
  // Contains functions and variables shared by VinciaShower and VinciaMatching

  class VinciaCommon {
    
  public:
    
    // Constructor.
    VinciaCommon() {isInitPtr=false; isInit=false;}

    // Destructor.
    virtual ~VinciaCommon() {}

    // Initialize pointers
    bool initPtr(ParticleData* particleDataPtrIn, Settings* settingsPtrIn) {
      particleDataPtr = particleDataPtrIn; 
      settingsPtr = settingsPtrIn; 
      isInitPtr=true;
      return true;
    }

    // Initialize data members
    bool init();

    // Get number of active flavors at a given Q scale
    int   getNf(double q) {
      if (q <= mc) return 3;
      else if (q <= mb) return 4;
      else if (q <= mt) return 5;
      else return 6;
    }

    // Verbosity level
    void setVerbose(int verboseIn) { verbose = verboseIn;}

    // Public data members: strong coupling in MSbar and CMW schemes
    AlphaStrong    alphaStrong;

  private:

    // Quark masses
    double ms, mc, mb, mt;

    // Pointers
    Settings*      settingsPtr; 
    ParticleData*  particleDataPtr;

    // Internal flags and settings
    bool isInitPtr, isInit;
    int verbose;

  };
 
  //**************************************************************************
  
  // Globally accessible functions and constants in the Vincia namespace
  
  // Color factors in Vincia normalization
  const double CA=3.0;
  const double CF=8.0/3.0;
  const double TR=1.0;
  const double NC=3.0;

  // Indices of antenna functions
  const int iQQemit  = 0;
  const int iQGemit  = 1;
  const int iGGemit  = 2;
  const int iQGsplit = 3;
  const int iGGsplit = 4;

  // Mathematical constants
  const double gammaE=0.577215664901532860606512090082402431042;
  
  // Max index for antenna coefficients
  const int LaurentIndexMax  = 10;
  // MR 10.7.2010: Max index for the invariants in the massive
  // antennae
  const int LaurentIndexMaxMassiveInv = 4;
  const int LaurentIndexMaxMassiveMass = 8;
  
  // A few useful auxiliary functions like extra invariants and dot products
  double m(const Vec4&);
  double m2(const Vec4&);
  double m2(const Vec4&, const Vec4&, const Vec4&);
  double m2(const Vec4&, const Vec4&, const Vec4&, const Vec4&);
  double m2(const Particle&, const Particle&, const Particle&);
  double dot4(const Vec4&, const Vec4&);
  double dot4(const Particle&, const Particle&);
  void Vec4toD4(const Vec4&, double*);
  void Vec4toMG(const Vec4&, double*);
  
  // Utilities to find and sort a bunch of partons
  int colorSort(Event&, vector<int>&);
  int findDaughters(Event&, int, vector<int>&);
  // Function to return whether parton i in event has iAncestor in its history
  bool isAncestor(Event& event, int i, int iAncestor);
  
  // Functors seem to be the only way to use zbrent for the rescaling 
  // variable needed in the massive Rambo version
  // TFunctor should be a template class, but the generality is 
  // not needed at this point
  
  // pure virtual function overloading operator()
  class TFunctor {
    public:
      virtual double operator()(double) =0;
  };
  
  class TXiFunctor : public TFunctor {
    public:
      // constructor takes a vector with the masses and one with the 
      // energies
      TXiFunctor(vector<double> mIn, vector<double> energiesIn);
      double operator() (double); 
    private:
      vector<double> m;
      vector<double> energies;
  };
  
  // derive a class which can wrap function pointers
  class TPtrFunctor : public TFunctor {
    public:
      TPtrFunctor(double (*fPtrIn)(double) ){ fPtr = fPtrIn; };
      double operator() (double arg) { return fPtr(arg); };
    private:
      // pointer to a function taking one double and returning one
      // double
      double (*fPtr)(double);
  };

  
  // Math support tools
  double zbrent(double (*)(double), double, double, double, double);
  // zbrentAlt is code duplication of zbrent, should replace zbrent, 
  // but testing is not complete
  // at present, doesn't work if the reference to TFunctor is declared
  // const
  double zbrentAlt(TFunctor&, double, double, double, double);
  double integrateGauss( double (*)(double), double, double, double);
  void makeSpline(vector<double>&, vector<double>&, vector<double>&);
  double getSpline(const double, vector<double>&, vector<double>&, 
		   vector<double>&);
  
  // Math support functions
  double Li2(const double,const double kmax=100.0,const double xerr=TINY);
  double lnGamma(const double);
  double Gamma(const double);
  int factorial(const int);
  int binomial(const int,int);
  
  // MR 21.12.2010: Gram determinant
  // the inviariants used in the argument are 2*pi*pj
  double gDet(
    double s12tilde, double s23tilde, double s13tilde, 
    double m1, double m2, double m3);
  // version with y13 omitted for normalised coordinates
  double gDet(
    double y12tilde, double y23tilde, 
    double mu1, double mu2,double mu3);
  double gDet(Vec4 p1, Vec4 p2, Vec4 p3);
    
  //**************************************************************************
  
  // A simple class to contain a spline
  
  class spline {
  public: 
    
    // constructor
    spline() {};
    
    // Members
    vector<double> x, y, d2;
  };
  
  //**************************************************************************

  // function to convert a vector of particle names to id's
  vector<int> nameVec2idVec(vector<string> nameVec);
  
  // From http://stackoverflow.com/questions/5343190
  inline string replaceString(string subject, const string& search,
                              const string& replace) {
    string::size_type pos = 0;
    while ((pos = subject.find(search, pos)) != string::npos) {
      subject.replace(pos, search.length(), replace);
      pos += replace.length();
    }
    return subject;
  }
  
  inline string sanitizeFileName(string fileName) {
    map<string, string> rep;
    rep["/"] = "_div_";
    rep[":"] = "_colon_";
    string retVal = fileName;
    for (map<string, string>::const_iterator it = rep.begin(); it != rep.end();
         ++it) {
      retVal = replaceString(retVal, it->first, it->second);
    }
    return retVal;
  }
  
} // End VINCIA namespace
#endif
